/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#ifndef __NTHW_TBH_H__
#define __NTHW_TBH_H__

/*
 * TX HB descriptor control
 */
typedef struct {
  uint8_t dt;        /* 0: PCAP descriptor. */
                     /* 1: Standard descriptor. */
                     /* 2: Dynamic descriptor */

  uint8_t ts_fmt;    /* Timestamp format used for the host buffer. */
                     /*  */
                     /* 0: PCAP-ns - 1ns units, start date: 1970-01-01. */
                     /* 1: Native - 10ns units, start date: 1970-01-01. */
                     /* 2: PCAP-us - 1us units, start date: 1970-01-01. */
                     /* 3: NDIS-native - 10ns units, start date: 1601-01-01. */
                     /* 4: Native-ns - 1ns units, start date: 1970-01-01. */

  uint8_t wl_use;    /* 0 : Stored length is used to determine frame size. */
                     /* 1 : Use wire length to determine frame size. */

  uint8_t fcs;       /* 0 : Frame stored in the host buffer is without FCS. */
                     /* 1 : Frame stored in the host buffer is with FCS. */
                     /*     Note that PCAP format usually does not include the FCS. */

  uint8_t ig_use;    /* 0 : TX ignore bit not used. */
                     /* 1 : TX ignore bit used to determine whether packet should be transmitted. */

  uint8_t ig_pos;    /* 
                      * Bit position in the packet descriptor of where the TX Ignore bit is located.
                      * 0 : is the lower bit of the byte 0;
                      * 7 : is the upper bit bit of byte 0;
                      * 8 : is the lower bit of the byte 1; etc.
                      */

  uint8_t pfd;       /* Determines of the TX destination port should be taken */
                     /* from the packet descriptor or from the fixed PORT field below. */
                     /* 0 : Fixed port as determined by PORT field below. */
                     /* 1 : Use port field from packet descriptor; PORT_POS defines where. */
                     /* In either case the port mask in the Host Buffer Ports memory */
                     /* must have a set bit for the possible port destinations. */

  uint8_t port;      /* The destination port to use when PFD=0. */

  uint8_t port_pos;  /* 
                      * Bit position in the packet descriptor of where the port is located when
                      * PFD=1;
                      * Position indicates the least significant bit of the port field.
                      */

  uint8_t port_mask; /* 
                      * The bit mask defines the allowed TX ports for thus host buffer.
                      * It is important that software sets this field to be consistent
                      * with the PORT_MAPPING register, i.e. if multiple bits are set in PORT_MASK
                      * the these ports must map to the same FIFO in the PORT_MAPPING register.
                      */

  /* TBH version >= 0.2 */
  uint8_t tx_now_use;
  uint8_t tx_now_pos;

  uint8_t tx_on_ts;

  uint8_t set_clock_use;
  uint8_t set_clock_pos;

  uint8_t ts_append;
} TbhHostBufferDescriptorControl_t;

#ifndef _NTSD_MAX_TX_PORTS_PER_ADAPTER_
#define _NTSD_MAX_TX_PORTS_PER_ADAPTER_ (8)
#endif

struct NthwTbh
{
  nthw_fpga_t* mpFpga;
  nthw_module_t* mpModTbh;
  int mnInstance;

  int mnModuleMajorVersion;
  int mnModuleMinorVersion;

  uint64_t mnModuleVersionPacked64;

  int mnTbhDebugDln;
  int mnTxPorts;
  uint8_t mPortCount;

  nthw_register_t* mpRegCtrl;
  nthw_field_t* mpFldCtrlEnable;

  nthw_register_t* mpRegCtrlSegment;
  nthw_field_t* mpFldCtrlSegmentSize;

  nthw_register_t* mpRegPortMapping;
  nthw_field_t* mpFldPortMapping[_NTSD_MAX_TX_PORTS_PER_ADAPTER_];

  nthw_register_t* mpRegHbInfoMemCtrl;
  nthw_field_t* mpFldHbInfoMemCtrlAdr;
  nthw_field_t* mpFldHbInfoMemCtrlCount;
  nthw_register_t* mpRegHbInfoMemData;
  nthw_field_t* mpFldHbInfoMemDataSize;

  nthw_register_t* mpRegHbPortMapMemCtrl;
  nthw_field_t* mpFldHbPortMapMemCtrlAdr;
  nthw_field_t* mpFldHbPortMapMemCtrlCount;
  nthw_register_t* mpRegHbPortMapMemData;
  nthw_field_t* mpFldHbPortMapMemDataMapping;

  nthw_register_t* mpRegDisable;
  nthw_field_t* mpFldDisableDisable;
  nthw_field_t* mpFldDisableHb;

  nthw_register_t* mpRegSetRdPointer;
  nthw_field_t* mpFldSetRdPointerHb;
  nthw_field_t* mpFldSetRdPointerOffset;

  nthw_register_t* mpRegStatus;
  nthw_field_t* mpFldStatusStopped;

  nthw_register_t* mpRegHbDscMemCtrl;
  nthw_field_t* mpFldHbDscMemCtrlAdr;
  nthw_field_t* mpFldHbDscMemCtrlCount;

  nthw_register_t* mpRegHbDscMemData;
  nthw_field_t* mpFldHbDscMemDataDt;
  nthw_field_t* mpFldHbDscMemDataWlUse;
  nthw_field_t* mpFldHbDscMemDataFcs;
  nthw_field_t* mpFldHbDscMemDataIgUse;
  nthw_field_t* mpFldHbDscMemDataIgPos;
  nthw_field_t* mpFldHbDscMemDataPfd;
  nthw_field_t* mpFldHbDscMemDataPort;
  nthw_field_t* mpFldHbDscMemDataPortPos;
  nthw_field_t* mpFldHbDscMemDataPortMask;

  /* TBH version >= 0.2 */
  nthw_field_t* mpFldHbDscMemDataTsFmt;
  nthw_field_t* mpFldHbDscMemDataTxNowUse;
  nthw_field_t* mpFldHbDscMemDataTxNowPos;
  nthw_field_t* mpFldHbDscMemDataTxOnTs;
  nthw_field_t* mpFldHbDscMemDataSetClockUse;
  nthw_field_t* mpFldHbDscMemDataSetClockPos;
  nthw_field_t* mpFldHbDscMemDataTsAppend;

  nthw_register_t* mpRegDbgDlnErr;
  nthw_field_t* mpFldDbgDlnErrE;
  nthw_field_t* mpFldDbgDlnErrHb;

  nthw_register_t* mpRegDbgIllegalRange;
  nthw_field_t* mpFldDbgIllegalRangeE;
  nthw_field_t* mpFldDbgIllegalRangeHb;
};

typedef struct NthwTbh NthwTbh_t;
typedef struct NthwTbh NthwTbh;

NthwTbh_t* NthwTbh_new(void);
void NthwTbh_delete(NthwTbh_t* p);
int NthwTbh_init(NthwTbh_t *p, nthw_fpga_t* pFpga, int nInstance);

void NthwTbh_enable(NthwTbh_t* p);
void NthwTbh_disable(NthwTbh_t* p);

uint32_t NthwTbh_getEnable(NthwTbh_t* p);
void NthwTbh_setEnable(NthwTbh_t* p, uint32_t value);

void NthwTbh_setSegmentSize(NthwTbh_t* p, uint32_t segmentSize);

void NthwTbh_setPortToFIFOMapping(NthwTbh_t* p, uint8_t* portToFifoMap, uint8_t portCount);

void NthwTbh_setHostBufferSize(NthwTbh_t* p, uint32_t nTxHbFeedNo, uint64_t hbPhysicalSize);

void NthwTbh_setHostBufferPortMask(NthwTbh_t* p, uint32_t nTxHbFeedNo, uint8_t portMask);

void NthwTbh_setHostBufferDescriptorControl(NthwTbh_t* p, uint32_t nTxHbFeedNo, TbhHostBufferDescriptorControl_t* hbControl);

int NthwTbh_setRdPointer(NthwTbh_t* p, const int nHbFeed, uint64_t nOffset, const bool bPauseHb);

int NthwTbh_getDebugDlnStatus(NthwTbh_t* p);

int NthwTbh_getDebugIllegalRangeStatus(NthwTbh_t* p);

int NthwTbh_CtrlHbFeed(NthwTbh_t* p, int nTxHbFeedNo, bool enable, bool bThrowException);

int NthwTbh_getLevel(NthwTbh_t* p);

#endif /* __NTHW_TBH_H__ */
