/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "nt_util.h"
#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"
#include "nthw_fpga_model.h"

#include "flow_nthw_info.h"

#include <stdlib.h> /* malloc */
#include <string.h> /* memset */

static inline unsigned int clamp_one(unsigned int val)
{
  return val > 1 ? 1 : val;
}

struct InfoNthw *InfoNthw_new(void)
{
  struct InfoNthw *p = malloc(sizeof(struct InfoNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void InfoNthw_delete(struct InfoNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int InfoNthw_init(struct InfoNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  assert(nInstance >= 0 && nInstance < 256);

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;

  unsigned int KmPresent  = clamp_one(nthw_fpga_get_product_param(pFpga, NT_KM_PRESENT, 0));
  unsigned int KccPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_CAT_KCC_PRESENT, 0));
  unsigned int IoaPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_IOA_PRESENT, 0));
  unsigned int RoaPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_ROA_PRESENT, 0));
  unsigned int DbsPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_DBS_PRESENT, 0));
  unsigned int FlmPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_FLM_PRESENT, 0));
  unsigned int HstPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_HST_PRESENT, 0));

  /* Modules for Tx Packet Edit function */
  unsigned int HfuPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_HFU_PRESENT, 0));
  unsigned int TxCpyPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_TX_CPY_PRESENT, 0));
  unsigned int TxInsPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_TX_INS_PRESENT, 0));
  unsigned int TxRplPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_TX_RPL_PRESENT, 0));
  unsigned int CsuPresent = clamp_one(nthw_fpga_get_product_param(pFpga, NT_CSU_PRESENT, 0));
  unsigned int TpePresent = (HfuPresent && TxCpyPresent && TxInsPresent && TxRplPresent && CsuPresent) ? 1 : 0;

  p->nPhyPorts = nthw_fpga_get_product_param(pFpga, NT_PHY_PORTS, 0);
  p->nRxPorts = nthw_fpga_get_product_param(pFpga, NT_RX_PORTS, 0);
  p->nLtxAvail = nthw_fpga_get_product_param(pFpga, NT_LR_PRESENT, 0);
  p->nbCatFunc = nthw_fpga_get_product_param(pFpga, NT_CAT_FUNCS, 0);
  p->nbCategories = nthw_fpga_get_product_param(pFpga, NT_CATEGORIES, 0);
  p->nbQueues = nthw_fpga_get_product_param(pFpga, NT_QUEUES, 0);
  p->nbFlowTypes = nthw_fpga_get_product_param(pFpga, NT_KM_FLOW_TYPES, 0) * clamp_one(KmPresent + FlmPresent);
  p->nbPmExt = nthw_fpga_get_product_param(pFpga, NT_CAT_N_EXT, 0);
  p->nbLen = nthw_fpga_get_product_param(pFpga, NT_CAT_N_LEN, 0);
  p->nbKccSize = nthw_fpga_get_product_param(pFpga, NT_CAT_KCC_SIZE, 0) * KccPresent;
  p->nbKccBanks = nthw_fpga_get_product_param(pFpga, NT_CAT_KCC_BANKS, 0) * KccPresent;
  p->nbKmCategories = nthw_fpga_get_product_param(pFpga, NT_KM_CATEGORIES, 0) * KmPresent;
  p->nbKmCamBanks = nthw_fpga_get_product_param(pFpga, NT_KM_CAM_BANKS, 0) * KmPresent;
  p->nbKmCamRecordWords = nthw_fpga_get_product_param(pFpga, NT_KM_CAM_REC_WORDS, 0) * KmPresent;
  p->nbKmCamRecords = nthw_fpga_get_product_param(pFpga, NT_KM_CAM_RECORDS, 0) * KmPresent;
  p->nbKmTcamBanks = nthw_fpga_get_product_param(pFpga, NT_KM_TCAM_BANKS, 0) * KmPresent;
  p->nbKmTcamBankWidth = nthw_fpga_get_product_param(pFpga, NT_KM_TCAM_BANK_WIDTH, 0) * KmPresent;
  p->nbFlmCategories = nthw_fpga_get_product_param(pFpga, NT_FLM_CATEGORIES, 0) * FlmPresent;
  p->nbFlmSizeMb = nthw_fpga_get_product_param(pFpga, NT_FLM_SIZE_MB, 0);
  p->nbFlmEntrySize = nthw_fpga_get_product_param(pFpga, NT_FLM_ENTRY_SIZE, 0);
  p->nbFlmVariant = nthw_fpga_get_product_param(pFpga, NT_FLM_VARIANT, 0);
  p->nbFlmPrios = nthw_fpga_get_product_param(pFpga, NT_FLM_PRIOS, 0) * FlmPresent;
  p->nbFlmPstProfiles = nthw_fpga_get_product_param(pFpga, NT_FLM_PST_PROFILES, 0) * FlmPresent;
  p->nbHstCategories = nthw_fpga_get_product_param(pFpga, NT_HST_CATEGORIES, 0) * HstPresent;
  p->nbQslCategories = nthw_fpga_get_product_param(pFpga, NT_QSL_CATEGORIES, 0);
  p->nbQslQstEntries = nthw_fpga_get_product_param(pFpga, NT_QSL_QST_SIZE, 0);
  p->nbPdbCategories = nthw_fpga_get_product_param(pFpga, NT_PDB_CATEGORIES, 0);
  p->nbIoaCategories = nthw_fpga_get_product_param(pFpga, NT_IOA_CATEGORIES, 0) * IoaPresent;
  p->nbRoaCategories = nthw_fpga_get_product_param(pFpga, NT_ROA_CATEGORIES, 0) * RoaPresent;
  p->nbDbsCategories = MIN(nthw_fpga_get_product_param(pFpga, NT_DBS_RX_QUEUES, 0), nthw_fpga_get_product_param(pFpga, NT_DBS_TX_QUEUES, 0)) * DbsPresent;
  p->nbCatKmIfCnt = nthw_fpga_get_product_param(pFpga, NT_CAT_KM_IF_CNT, KmPresent + FlmPresent);
  p->mCatKmIfM0 = nthw_fpga_get_product_param(pFpga, NT_CAT_KM_IF_M0, -1);
  p->mCatKmIfM1 = nthw_fpga_get_product_param(pFpga, NT_CAT_KM_IF_M1, -1);
  p->nbTpeCategories = nthw_fpga_get_product_param(pFpga, NT_TPE_CATEGORIES, 0) * TpePresent;
  p->nbTxCpyWriters = nthw_fpga_get_product_param(pFpga, NT_TX_CPY_WRITERS, 0) * TpePresent;
  p->nbTxCpyMaskMem = nthw_fpga_get_product_param(pFpga, NT_CPY_MASK_MEM, 0) * TpePresent;
  p->nbTxRplDepth = nthw_fpga_get_product_param(pFpga, NT_TX_RPL_DEPTH, 0) * TpePresent;
  p->nbTxRplExtCategories = nthw_fpga_get_product_param(pFpga, NT_TX_RPL_EXT_CATEGORIES, 0) * TpePresent;
  p->nbTpeIfrCategories = nthw_fpga_get_product_param(pFpga, NT_TX_MTU_PROFILE_IFR, 0);
  return 0;
}

unsigned int InfoNthw_getNbPhyPorts(const struct InfoNthw *p)
{
  return p->nPhyPorts;
}

unsigned int InfoNthw_getNbRxPorts(const struct InfoNthw *p)
{
  return p->nRxPorts;
}

unsigned int InfoNthw_getLtxAvail(const struct InfoNthw *p)
{
  return p->nLtxAvail;
}

unsigned int InfoNthw_getNbCategories(const struct InfoNthw *p)
{
  return p->nbCategories;
}

unsigned int InfoNthw_getKccSize(const struct InfoNthw *p)
{
  return p->nbKccSize;
}

unsigned int InfoNthw_getKccBanks(const struct InfoNthw *p)
{
  return p->nbKccBanks;
}

unsigned int InfoNthw_getNbQueues(const struct InfoNthw *p)
{
  return p->nbQueues;
}

unsigned int InfoNthw_getNbCatFuncs(const struct InfoNthw *p)
{
  return p->nbCatFunc;
}

unsigned int InfoNthw_getNbKMFlowTypes(const struct InfoNthw *p)
{
  return p->nbFlowTypes;
}

unsigned int InfoNthw_getNbPmExt(const struct InfoNthw *p)
{
  return p->nbPmExt;
}

unsigned int InfoNthw_getNbLen(const struct InfoNthw *p)
{
  return p->nbLen;
}

unsigned int InfoNthw_getNbKmCategories(const struct InfoNthw *p)
{
  return p->nbKmCategories;
}

unsigned int InfoNthw_getNbKmCamBanks(const struct InfoNthw *p)
{
  return p->nbKmCamBanks;
}

unsigned int InfoNthw_getNbKmCamRecordWords(const struct InfoNthw *p)
{
  return p->nbKmCamRecordWords;
}

unsigned int InfoNthw_getNbKmCamRecords(const struct InfoNthw *p)
{
  return p->nbKmCamRecords;
}

unsigned int InfoNthw_getNbKmTcamBanks(const struct InfoNthw *p)
{
  return p->nbKmTcamBanks;
}

unsigned int InfoNthw_getNbKmTcamBankWidth(const struct InfoNthw *p)
{
  return p->nbKmTcamBankWidth;
}

unsigned int InfoNthw_getNbFlmCategories(const struct InfoNthw *p)
{
  return p->nbFlmCategories;
}

unsigned int InfoNthw_getNbFlmSizeMb(const struct InfoNthw *p)
{
  return p->nbFlmSizeMb;
}

unsigned int InfoNthw_getNbFlmEntrySize(const struct InfoNthw *p)
{
  return p->nbFlmEntrySize;
}

unsigned int InfoNthw_getNbFlmVariant(const struct InfoNthw *p)
{
  return p->nbFlmVariant;
}

unsigned int InfoNthw_getNbFlmPrios(const struct InfoNthw *p)
{
  return p->nbFlmPrios;
}

unsigned int InfoNthw_getNbFlmPstProfiles(const struct InfoNthw *p)
{
  return p->nbFlmPstProfiles;
}

unsigned int InfoNthw_getNbHstCategories(const struct InfoNthw *p)
{
  return p->nbHstCategories;
}

unsigned int InfoNthw_getNbQslCategories(const struct InfoNthw *p)
{
  return p->nbQslCategories;
}

unsigned int InfoNthw_getNbQslQstEntries(const struct InfoNthw *p)
{
  return p->nbQslQstEntries;
}

unsigned int InfoNthw_getNbPdbCategories(const struct InfoNthw *p)
{
  return p->nbPdbCategories;
}

unsigned int InfoNthw_getNbIoaCategories(const struct InfoNthw *p)
{
  return p->nbIoaCategories;
}

unsigned int InfoNthw_getNbRoaCategories(const struct InfoNthw *p)
{
  return p->nbRoaCategories;
}

unsigned int InfoNthw_getNbDbsCategories(const struct InfoNthw *p)
{
  return p->nbDbsCategories;
}

unsigned int InfoNthw_getNbCatKmIfCnt(const struct InfoNthw *p)
{
  return p->nbCatKmIfCnt;
}

unsigned int InfoNthw_getNbCatKmIfM0(const struct InfoNthw *p)
{
  return p->mCatKmIfM0;
}

unsigned int InfoNthw_getNbCatKmIfM1(const struct InfoNthw *p)
{
  return p->mCatKmIfM1;
}

unsigned int InfoNthw_getNbTpeCategories(const struct InfoNthw *p)
{
  return p->nbTpeCategories;
}

unsigned int InfoNthw_getNbTxCpyWriters(const struct InfoNthw *p)
{
  return p->nbTxCpyWriters;
}

unsigned int InfoNthw_getNbTxCpyMaskMem(const struct InfoNthw *p)
{
  return p->nbTxCpyMaskMem;
}

unsigned int InfoNthw_getNbTxRplDepth(const struct InfoNthw *p)
{
  return p->nbTxRplDepth;
}

unsigned int InfoNthw_getNbTxRplExtCategories(const struct InfoNthw *p)
{
  return p->nbTxRplExtCategories;
}

unsigned int InfoNthw_getNbTpeIfrCategories(const struct InfoNthw *p)
{
  return p->nbTpeIfrCategories;
}
