/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_roa.h"

#include <stdlib.h> /* malloc */
#include <string.h> /* memset */

void roaNthw_setDebugMode(struct RoaNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mRoa, nDebugMode);
}

struct RoaNthw *RoaNthw_new(void)
{
  struct RoaNthw *p = malloc(sizeof(struct RoaNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void RoaNthw_delete(struct RoaNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int RoaNthw_init(struct RoaNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = Fpga_queryModule(pFpga, MOD_ROA, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: ROA %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mRoa = pMod;

#if defined(DEBUG) && (0)
  nthw_module_set_debug_mode(p->mRoa, 0xFF); /* DEBUG: */
#endif

  /* TUN HDR */
  p->mpTunHdrCtrl = nthw_module_get_register(p->mRoa, ROA_TUNHDR_CTRL);
  p->mpTunHdrAddr = nthw_register_get_field(p->mpTunHdrCtrl, ROA_TUNHDR_CTRL_ADR);
  p->mpTunHdrCnt = nthw_register_get_field(p->mpTunHdrCtrl, ROA_TUNHDR_CTRL_CNT);
  p->mpTunHdrData = nthw_module_get_register(p->mRoa, ROA_TUNHDR_DATA);
  p->mpTunHdrDataTunnelHdr =     nthw_register_get_field(p->mpTunHdrData, ROA_TUNHDR_DATA_TUNNEL_HDR);
  /* TUN CFG */
  p->mpTunCfgCtrl = nthw_module_get_register(p->mRoa, ROA_TUNCFG_CTRL);
  p->mpTunCfgAddr = nthw_register_get_field(p->mpTunCfgCtrl, ROA_TUNCFG_CTRL_ADR);
  p->mpTunCfgCnt = nthw_register_get_field(p->mpTunCfgCtrl, ROA_TUNCFG_CTRL_CNT);
  p->mpTunCfgData = nthw_module_get_register(p->mRoa, ROA_TUNCFG_DATA);
  p->mpTunCfgDataTunLen = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_LEN);
  p->mpTunCfgDataTunType = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_TYPE);
  p->mpTunCfgDataTunVlan = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_VLAN);
  p->mpTunCfgDataIpType = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IP_TYPE);
  p->mpTunCfgDataIpcsUpd = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IPCS_UPD);
  p->mpTunCfgDataIpcsPrecalc = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IPCS_PRECALC);
  p->mpTunCfgDataIptlUpd = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IPTL_UPD);
  p->mpTunCfgDataIptlPrecalc = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_IPTL_PRECALC);
  p->mpTunCfgDataVxlanUdpLenUpd = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TUN_VXLAN_UDP_LEN_UPD);
  p->mpTunCfgDataTxLagIx = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_TX_LAG_IX);
  p->mpTunCfgDataRecirculate = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_RECIRCULATE);
  p->mpTunCfgDataPushTunnel = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_PUSH_TUNNEL);
  p->mpTunCfgDataRecircPort = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_RECIRC_PORT);
  p->mpTunCfgDataRecircBypass = nthw_register_get_field(p->mpTunCfgData, ROA_TUNCFG_DATA_RECIRC_BYPASS);
  /* CONFIG */
  p->mpConfig = nthw_module_get_register(p->mRoa, ROA_CONFIG);
  p->mpConfigFwdRecirculate = nthw_register_get_field(p->mpConfig, ROA_CONFIG_FWD_RECIRCULATE);
  p->mpConfigFwdNormalPcks = nthw_register_get_field(p->mpConfig, ROA_CONFIG_FWD_NORMAL_PCKS);
  p->mpConfigFwdTxPort0 = nthw_register_get_field(p->mpConfig, ROA_CONFIG_FWD_TXPORT0);
  p->mpConfigFwdTxPort1 = nthw_register_get_field(p->mpConfig, ROA_CONFIG_FWD_TXPORT1);
  p->mpConfigFwdCellBuilderPcks = nthw_register_get_field(p->mpConfig, ROA_CONFIG_FWD_CELLBUILDER_PCKS);
  p->mpConfigFwdNonNormalPcks = nthw_register_get_field(p->mpConfig, ROA_CONFIG_FWD_NON_NORMAL_PCKS);
  /* LAG */
  p->mpLagCfgCtrl = nthw_module_get_register(p->mRoa, ROA_LAGCFG_CTRL);
  p->mpLagCfgAddr = nthw_register_get_field(p->mpLagCfgCtrl, ROA_LAGCFG_CTRL_ADR);
  p->mpLagCfgCnt = nthw_register_get_field(p->mpLagCfgCtrl, ROA_LAGCFG_CTRL_CNT);
  p->mpLagCfgData = nthw_module_get_register(p->mRoa, ROA_LAGCFG_DATA);
  p->mpLagCfgDataTxPhyPort = nthw_register_get_field(p->mpLagCfgData, ROA_LAGCFG_DATA_TXPHY_PORT);

  return 0;
}

/* TUN HDR */
void RoaNthw_TunHdrSelect(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunHdrAddr, val);
}

void RoaNthw_TunHdrCnt(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunHdrCnt, val);
}

void RoaNthw_TunHdrTunnelHdr(const struct RoaNthw *p, uint32_t *val)
{
  nthw_field_set_val(p->mpTunHdrDataTunnelHdr, val, 4);
}

void RoaNthw_TunHdrFlush(const struct RoaNthw *p)
{
  nthw_register_flush(p->mpTunHdrCtrl, 1);
  nthw_register_flush(p->mpTunHdrData, 1);
}

/* TUN CFG */
void RoaNthw_TunCfgSelect(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgAddr, val);
}

void RoaNthw_TunCfgCnt(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgCnt, val);
}

void RoaNthw_TunCfgTunLen(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataTunLen, val);
}

void RoaNthw_TunCfgTunType(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataTunType, val);
}

void RoaNthw_TunCfgTunVlan(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataTunVlan, val);
}

void RoaNthw_TunCfgIpType(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataIpType, val);
}

void RoaNthw_TunCfgIpcsUpd(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataIpcsUpd, val);
}

void RoaNthw_TunCfgIpcsPrecalc(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataIpcsPrecalc, val);
}

void RoaNthw_TunCfgIptlUpd(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataIptlUpd, val);
}

void RoaNthw_TunCfgIptlPrecalc(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataIptlPrecalc, val);
}

void RoaNthw_TunCfgVxlanUdpLenUpd(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataVxlanUdpLenUpd, val);
}

void RoaNthw_TunCfgTxLagIx(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataTxLagIx, val);
};

void RoaNthw_TunCfgRecirculate(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataRecirculate, val);
}

void RoaNthw_TunCfgPushTunnel(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataPushTunnel, val);
}

void RoaNthw_TunCfgRecircPort(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataRecircPort, val);
}

void RoaNthw_TunCfgRecircBypass(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpTunCfgDataRecircBypass, val);
}

void RoaNthw_TunCfgFlush(const struct RoaNthw *p)
{
  nthw_register_flush(p->mpTunCfgCtrl, 1);
  nthw_register_flush(p->mpTunCfgData, 1);
}

/* ROA CONFIG */
void RoaNthw_ConfigFwdRecirculate(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpConfigFwdRecirculate, val);
}

void RoaNthw_ConfigFwdNormalPcks(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpConfigFwdNormalPcks, val);
}

void RoaNthw_ConfigFwdTxPort0(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpConfigFwdTxPort0, val);
}

void RoaNthw_ConfigFwdTxPort1(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpConfigFwdTxPort1, val);
}

void RoaNthw_ConfigFwdCellBuilderPcks(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpConfigFwdCellBuilderPcks, val);
}

void RoaNthw_ConfigFwdNonNormalPcks(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpConfigFwdNonNormalPcks, val);
}

void RoaNthw_ConfigFlush(const struct RoaNthw *p)
{
  nthw_register_flush(p->mpConfig, 1);
}

/* LAG */
void RoaNthw_LagCfgSelect(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpLagCfgAddr, val);
}

void RoaNthw_LagCfgCnt(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpLagCfgCnt, val);
}

void RoaNthw_LagCfgTxPhyPort(const struct RoaNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpLagCfgDataTxPhyPort, val);
}

void RoaNthw_LagCfgFlush(const struct RoaNthw *p)
{
  nthw_register_flush(p->mpLagCfgCtrl, 1);
  nthw_register_flush(p->mpLagCfgData, 1);
}
