/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"
#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_rpp_lr.h"

#include <stdlib.h>
#include <string.h>

void rppLrNthw_setDebugMode(struct RppLrNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mRppLr, nDebugMode);
}

struct RppLrNthw *RppLrNthw_new(void)
{
  struct RppLrNthw *p = malloc(sizeof(struct RppLrNthw));
  if (p) {
    (void)memset(p, 0, sizeof(*p));
  }
  return p;
}

void RppLrNthw_delete(struct RppLrNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int RppLrNthw_init(struct RppLrNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = Fpga_queryModule(pFpga, MOD_RPP_LR, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: RppLr %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mRppLr = Fpga_queryModule(pFpga, MOD_RPP_LR, nInstance);

  p->mpRcpCtrl = nthw_module_get_register(p->mRppLr, RPP_LR_RCP_CTRL);
  p->mpRcpAddr = nthw_register_get_field(p->mpRcpCtrl, RPP_LR_RCP_CTRL_ADR);
  p->mpRcpCnt = nthw_register_get_field(p->mpRcpCtrl, RPP_LR_RCP_CTRL_CNT);
  p->mpRcpData = nthw_module_get_register(p->mRppLr, RPP_LR_RCP_DATA);
  p->mpRcpDataExp = nthw_register_get_field(p->mpRcpData, RPP_LR_RCP_DATA_EXP);

  p->mpIfrRcpCtrl = nthw_module_query_register(p->mRppLr, RPP_LR_IFR_RCP_CTRL);
  p->mpIfrRcpAddr = nthw_register_query_field(p->mpIfrRcpCtrl, RPP_LR_IFR_RCP_CTRL_ADR);
  p->mpIfrRcpCnt = nthw_register_query_field(p->mpIfrRcpCtrl, RPP_LR_IFR_RCP_CTRL_CNT);
  p->mpIfrRcpData = nthw_module_query_register(p->mRppLr, RPP_LR_IFR_RCP_DATA);
  p->mpIfrRcpDataEn = nthw_register_query_field(p->mpIfrRcpData, RPP_LR_IFR_RCP_DATA_EN);
  p->mpIfrRcpDataMtu = nthw_register_query_field(p->mpIfrRcpData, RPP_LR_IFR_RCP_DATA_MTU);

  return 0;
}

void RppLrNthw_RcpSelect(const struct RppLrNthw *p, uint32_t val)
{
  assert(p->mpRcpAddr);
  nthw_field_set_val32(p->mpRcpAddr, val);
}

void RppLrNthw_RcpCnt(const struct RppLrNthw *p, uint32_t val)
{
  assert(p->mpRcpCnt);
  nthw_field_set_val32(p->mpRcpCnt, val);
}

void RppLrNthw_RcpExp(const struct RppLrNthw *p, uint32_t val)
{
  assert(p->mpRcpDataExp);
  nthw_field_set_val32(p->mpRcpDataExp, val);
}

void RppLrNthw_RcpFlush(const struct RppLrNthw *p)
{
  assert(p->mpRcpCtrl);
  assert(p->mpRcpData);
  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);
}

void RppLrNthw_IfrRcpSelect(const struct RppLrNthw *p, uint32_t val)
{
  assert(p->mpIfrRcpAddr);
  nthw_field_set_val32(p->mpIfrRcpAddr, val);
}

void RppLrNthw_IfrRcpCnt(const struct RppLrNthw *p, uint32_t val)
{
  assert(p->mpIfrRcpCnt);
  nthw_field_set_val32(p->mpIfrRcpCnt, val);
}

void RppLrNthw_IfrRcpEn(const struct RppLrNthw *p, uint32_t val)
{
  assert(p->mpIfrRcpDataEn);
  nthw_field_set_val32(p->mpIfrRcpDataEn, val);
}

void RppLrNthw_IfrRcpMtu(const struct RppLrNthw *p, uint32_t val)
{
  assert(p->mpIfrRcpDataMtu);
  nthw_field_set_val32(p->mpIfrRcpDataMtu, val);
}

void RppLrNthw_IfrRcpFlush(const struct RppLrNthw *p)
{
  assert(p->mpIfrRcpCtrl);
  assert(p->mpIfrRcpData);
  nthw_register_flush(p->mpIfrRcpCtrl, 1);
  nthw_register_flush(p->mpIfrRcpData, 1);
}
