/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"

#include "flow_nthw_slc_lr.h"

#include <stdlib.h> /* malloc */
#include <string.h> /* memset */

void slcLrNthw_setDebugMode(struct SlcLrNthw *p, unsigned int nDebugMode)
{
  nthw_module_set_debug_mode(p->mSlcLr, nDebugMode);
}

struct SlcLrNthw *SlcLrNthw_new(void)
{
  struct SlcLrNthw *p = malloc(sizeof(struct SlcLrNthw));
  if (p)
    (void)memset(p, 0, sizeof(*p));

  return p;
}

void SlcLrNthw_delete(struct SlcLrNthw *p)
{
  if (p) {
    (void)memset(p, 0, sizeof(*p));
    free(p);
  }
}

int SlcLrNthw_init(struct SlcLrNthw *p, nthw_fpga_t *pFpga, int nInstance)
{
  const char *const pAdapterIdStr = pFpga->p_fpga_info->mpAdapterIdStr;
  nthw_module_t* pMod = Fpga_queryModule(pFpga, MOD_SLC_LR, nInstance);

  assert(nInstance >= 0 && nInstance < 256);

  if (p == NULL) {
    return pMod == NULL ? -1 : 0;
  }

  if (pMod == NULL) {
    NT_LOG(ERR, NTHW, "%s: Slc %d: no such instance\n", pAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mPhysicalAdapterNo = (uint8_t)nInstance;
  p->mSlcLr = Fpga_queryModule(pFpga, MOD_SLC_LR, nInstance);

  /* RCP */
  p->mpRcpCtrl = nthw_module_get_register(p->mSlcLr, SLC_RCP_CTRL);
  p->mpRcpAddr = nthw_register_get_field(p->mpRcpCtrl, SLC_RCP_CTRL_ADR);
  p->mpRcpCnt = nthw_register_get_field(p->mpRcpCtrl, SLC_RCP_CTRL_CNT);
  p->mpRcpData = nthw_module_get_register(p->mSlcLr, SLC_RCP_DATA);
  p->mpRcpDataTailSlcEn = nthw_register_get_field(p->mpRcpData, SLC_RCP_DATA_TAIL_SLC_EN);
  p->mpRcpDataTailDyn = nthw_register_get_field(p->mpRcpData, SLC_RCP_DATA_TAIL_DYN);
  p->mpRcpDataTailOfs = nthw_register_get_field(p->mpRcpData, SLC_RCP_DATA_TAIL_OFS);
  p->mpRcpDataPcap = nthw_register_get_field(p->mpRcpData, SLC_RCP_DATA_PCAP);

  return 0;
}

/* RCP */
void SlcLrNthw_RcpSelect(const struct SlcLrNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpAddr, val);
}

void SlcLrNthw_RcpCnt(const struct SlcLrNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpCnt, val);
}

void SlcLrNthw_RcpTailSlcEn(const struct SlcLrNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTailSlcEn, val);
}

void SlcLrNthw_RcpTailDyn(const struct SlcLrNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTailDyn, val);
}

void SlcLrNthw_RcpTailOfs(const struct SlcLrNthw *p, int32_t val)
{
  nthw_field_set_val32(p->mpRcpDataTailOfs, val);
}

void SlcLrNthw_RcpPcap(const struct SlcLrNthw *p, uint32_t val)
{
  nthw_field_set_val32(p->mpRcpDataPcap, val);
}

void SlcLrNthw_RcpFlush(const struct SlcLrNthw *p)
{
  nthw_register_flush(p->mpRcpCtrl, 1);
  nthw_register_flush(p->mpRcpData, 1);
}
