/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"

#include "nthw_epp.h"

#include <errno.h>  /* ENOTSUP */

NthwEpp_t* NthwEpp_new(void)
{
  NthwEpp_t* p = malloc(sizeof(NthwEpp_t));
  if (p) {
    memset(p, 0, sizeof(NthwEpp_t));
  }
  return p;
}

void NthwEpp_delete(NthwEpp_t *p)
{
  if (p) {
    memset(p, 0, sizeof(NthwEpp_t));
    free(p);
  }
}

int NthwEpp_present(nthw_fpga_t* pFpga, int nInstance)
{
  return NthwEpp_init(NULL, pFpga, nInstance) == 0;
}

int NthwEpp_init(NthwEpp_t *p, nthw_fpga_t* pFpga, int nInstance)
{
  nthw_module_t* mod = Fpga_queryModule(pFpga, MOD_EPP, nInstance);

  if (p == NULL)
    return mod == NULL ? -1 : 0;

  if (mod == NULL) {
    NT_LOG(ERR, NTHW, "%s: EPP %d: no such instance\n", pFpga->p_fpga_info->mpAdapterIdStr,
        nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mnInstance = nInstance;
  p->mpModEpp = mod;

#if defined(DEBUG) && (0)
  nthw_module_set_debug_mode(p->mpModEpp, 0x00); /* DEBUG: */
#endif

  p->mnEppCategories = nthw_fpga_get_product_param(pFpga, NT_EPP_CATEGORIES, 0);

  p->mpRegReciepeMemoryControl = nthw_module_get_register(p->mpModEpp, EPP_RCP_CTRL);
  p->mpFldReciepeMemoryControlAdr = nthw_register_get_field(p->mpRegReciepeMemoryControl, EPP_RCP_CTRL_ADR);
  p->mpFldReciepeMemoryControlCnt = nthw_register_get_field(p->mpRegReciepeMemoryControl, EPP_RCP_CTRL_CNT);

  p->mpRegReciepeMemoryData = nthw_module_get_register(p->mpModEpp, EPP_RCP_DATA);
  p->mpFldReciepeMemoryDataTxMtuEppEnable  = nthw_register_get_field(p->mpRegReciepeMemoryData, EPP_RCP_DATA_TX_MTU_EPP_EN);
  p->mpFldReciepeMemoryDataQueueMtuEppEnable  = nthw_register_get_field(p->mpRegReciepeMemoryData, EPP_RCP_DATA_QUEUE_MTU_EPP_EN);
  p->mpFldReciepeMemoryDataSizeAdjustTxPort = nthw_register_get_field(p->mpRegReciepeMemoryData, EPP_RCP_DATA_SIZE_ADJUST_TXP);
  p->mpFldReciepeMemoryDataSizeAdjustVirtualPort = nthw_register_get_field(p->mpRegReciepeMemoryData, EPP_RCP_DATA_SIZE_ADJUST_VPORT);
  p->mpFldReciepeMemoryDataFixed18bL2Mtu = nthw_register_get_field(p->mpRegReciepeMemoryData, EPP_RCP_DATA_FIXED_18B_L2_MTU);
  p->mpFldReciepeMemoryDataTxpQosEppEnable  = nthw_register_get_field(p->mpRegReciepeMemoryData, EPP_RCP_DATA_TX_QOS_EPP_EN);
  p->mpFldReciepeMemoryDataQueueQosEppEnable  = nthw_register_get_field(p->mpRegReciepeMemoryData, EPP_RCP_DATA_QUEUE_QOS_EPP_EN);

  p->mpRegTxpPortMtuControl = nthw_module_get_register(p->mpModEpp, EPP_TXP_MTU_CTRL);
  p->mpFldTxpPortMtuControlAdr = nthw_register_get_field(p->mpRegTxpPortMtuControl, EPP_TXP_MTU_CTRL_ADR);
  p->mpFldTxpPortMtuControlCnt = nthw_register_get_field(p->mpRegTxpPortMtuControl, EPP_TXP_MTU_CTRL_CNT);

  p->mpRegTxpPortMtuData = nthw_module_get_register(p->mpModEpp, EPP_TXP_MTU_DATA);
  p->mpFldTxpPortMtuDataMaxMtu = nthw_register_get_field(p->mpRegTxpPortMtuData, EPP_TXP_MTU_DATA_MAX_MTU);

  p->mpRegQueueMtuControl = nthw_module_get_register(p->mpModEpp, EPP_QUEUE_MTU_CTRL);
  p->mpFldQueueMtuControlAdr = nthw_register_get_field(p->mpRegQueueMtuControl, EPP_QUEUE_MTU_CTRL_ADR);
  p->mpFldQueueMtuControlCnt = nthw_register_get_field(p->mpRegQueueMtuControl, EPP_QUEUE_MTU_CTRL_CNT);

  p->mpRegQueueMtuData = nthw_module_get_register(p->mpModEpp, EPP_QUEUE_MTU_DATA);
  p->mpFldQueueMtuDataMaxMtu = nthw_register_get_field(p->mpRegQueueMtuData, EPP_QUEUE_MTU_DATA_MAX_MTU);

  p->mpRegTxpQosControl = nthw_module_get_register(p->mpModEpp, EPP_TXP_QOS_CTRL);
  p->mpFldTxpQosControlAdr = nthw_register_get_field(p->mpRegTxpQosControl, EPP_TXP_QOS_CTRL_ADR);
  p->mpFldTxpQosControlCnt = nthw_register_get_field(p->mpRegTxpQosControl, EPP_TXP_QOS_CTRL_CNT);

  p->mpRegTxpQosData = nthw_module_get_register(p->mpModEpp, EPP_TXP_QOS_DATA);
  p->mpFldTxpQosDataEnable = nthw_register_get_field(p->mpRegTxpQosData, EPP_TXP_QOS_DATA_EN);
  p->mpFldTxpQosDataInformationRate = nthw_register_get_field(p->mpRegTxpQosData, EPP_TXP_QOS_DATA_IR);
  p->mpFldTxpQosDataInformationRateFractional = nthw_register_get_field(p->mpRegTxpQosData, EPP_TXP_QOS_DATA_IR_FRACTION);
  p->mpFldTxpQosDataBurstSize = nthw_register_get_field(p->mpRegTxpQosData, EPP_TXP_QOS_DATA_BS);

  p->mpRegVportQosControl = nthw_module_get_register(p->mpModEpp, EPP_VPORT_QOS_CTRL);
  p->mpFldVportQosControlAdr = nthw_register_get_field(p->mpRegVportQosControl, EPP_VPORT_QOS_CTRL_ADR);
  p->mpFldVportQosControlCnt = nthw_register_get_field(p->mpRegVportQosControl, EPP_VPORT_QOS_CTRL_CNT);

  p->mpRegVportQosData = nthw_module_get_register(p->mpModEpp, EPP_VPORT_QOS_DATA);
  p->mpFldVportQosDataEnable = nthw_register_get_field(p->mpRegVportQosData, EPP_VPORT_QOS_DATA_EN);
  p->mpFldVportQosDataInformationRate = nthw_register_get_field(p->mpRegVportQosData, EPP_VPORT_QOS_DATA_IR);
  p->mpFldVportQosDataInformationRateFractional = nthw_register_get_field(p->mpRegVportQosData, EPP_VPORT_QOS_DATA_IR_FRACTION);
  p->mpFldVportQosDataBurstSize = nthw_register_get_field(p->mpRegVportQosData, EPP_VPORT_QOS_DATA_BS);

  p->mpRegQueueVportControl = nthw_module_get_register(p->mpModEpp, EPP_QUEUE_VPORT_CTRL);
  p->mpFldQueueVportControlAdr = nthw_register_get_field(p->mpRegQueueVportControl, EPP_QUEUE_VPORT_CTRL_ADR);
  p->mpFldQueueVportControlCnt = nthw_register_get_field(p->mpRegQueueVportControl, EPP_QUEUE_VPORT_CTRL_CNT);

  p->mpRegQueueVportData = nthw_module_get_register(p->mpModEpp, EPP_QUEUE_VPORT_DATA);
  p->mpFldQueueVportDataVport = nthw_register_get_field(p->mpRegQueueVportData, EPP_QUEUE_VPORT_DATA_VPORT);

  return 0;
}

int NthwEpp_setup(NthwEpp_t *p)
{
  if(p == NULL)
    return 0;

  /* Set recieps for 2 first records */
  nthw_field_set_val32(p->mpFldReciepeMemoryControlCnt, 1);

  /* Zero all categories */
  for(int i = 0; i < p->mnEppCategories; ++i) {
    nthw_field_set_val32(p->mpFldReciepeMemoryControlAdr, i);
    nthw_register_flush(p->mpRegReciepeMemoryControl, 1);

    nthw_field_set_val32(p->mpFldReciepeMemoryDataTxMtuEppEnable, 0);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataQueueMtuEppEnable, 0);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataSizeAdjustTxPort, 0);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataSizeAdjustVirtualPort, 0);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataFixed18bL2Mtu, 0);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataTxpQosEppEnable, 0);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataQueueQosEppEnable, 0);
    nthw_register_flush(p->mpRegReciepeMemoryData, 1);
  }

  for(int i = 0; i < NRECIPE; ++i) {
    nthw_field_set_val32(p->mpFldReciepeMemoryControlAdr, i);
    nthw_register_flush(p->mpRegReciepeMemoryControl, 1);

    nthw_field_set_val32(p->mpFldReciepeMemoryDataTxMtuEppEnable, 1);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataQueueMtuEppEnable, 1);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataSizeAdjustTxPort, rcpDataSizeAdjustTxp[i]);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataSizeAdjustVirtualPort, rcpDataSizeAdjustVport[i]);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataFixed18bL2Mtu, 1);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataTxpQosEppEnable, 1);
    nthw_field_set_val32(p->mpFldReciepeMemoryDataQueueQosEppEnable, 1);
    nthw_register_flush(p->mpRegReciepeMemoryData, 1);
  }
  /* phy mtu setup */
  nthw_field_set_val32(p->mpFldTxpPortMtuControlCnt, 1);
  for(int i = 0; i < 2; ++i) {
    nthw_field_set_val32(p->mpFldTxpPortMtuControlAdr, i);
    nthw_register_flush(p->mpRegTxpPortMtuControl, 1);

    nthw_field_set_val32(p->mpFldTxpPortMtuDataMaxMtu, MTUINITVAL);
    nthw_register_flush(p->mpRegTxpPortMtuData, 1);
  }
  /* phy QoS setup */
  nthw_field_set_val32(p->mpFldTxpQosControlCnt, 1);
  for(int i = 0; i < 2; ++i) {
    nthw_field_set_val32(p->mpFldTxpQosControlAdr, i);
    nthw_register_flush(p->mpRegTxpQosControl, 1);

    nthw_field_set_val32(p->mpFldTxpQosDataEnable, 0);
    nthw_register_flush(p->mpRegTxpQosData, 1);
  }

  /* virt mtu setup */
  nthw_field_set_val32(p->mpFldQueueMtuControlCnt, 1);
  for(int i = 0; i < 128; ++i) {
    nthw_field_set_val32(p->mpFldQueueMtuControlAdr, i);
    nthw_register_flush(p->mpRegQueueMtuControl, 1);

    nthw_field_set_val32(p->mpFldQueueMtuDataMaxMtu, MTUINITVAL);
    nthw_register_flush(p->mpRegQueueMtuData, 1);
  }

  /* virt QoS setup */
  nthw_field_set_val32(p->mpFldVportQosControlCnt, 1);
  for(int i = 0; i < 128; ++i) {
    nthw_field_set_val32(p->mpFldVportQosControlAdr, i);
    nthw_register_flush(p->mpRegVportQosControl, 1);

    nthw_field_set_val32(p->mpFldVportQosDataEnable, 0);
    nthw_register_flush(p->mpRegVportQosData, 1);
  }

  return 0;
}

/*
 * Set the MTU registers in context with the current setMTU request.
 */
int NthwEpp_setMtu(NthwEpp_t *p, uint32_t port, uint32_t maxMtu, NtMetaPortType_t portType)
{
  if(p == NULL)
    return 0;

  if (portType == PORT_TYPE_PHYSICAL) {
    /* Set the TXP Mtu control register */
    nthw_field_set_val32(p->mpFldTxpPortMtuControlAdr, port);
    nthw_field_set_val32(p->mpFldTxpPortMtuControlCnt, 1);
    nthw_register_flush(p->mpRegTxpPortMtuControl, 1);

    /* Set the TXP Mtu control register */
    nthw_field_set_val32(p->mpFldTxpPortMtuDataMaxMtu, maxMtu);
    nthw_register_flush(p->mpRegTxpPortMtuData, 1);
  } else if (portType == PORT_TYPE_VIRTUAL) {
    /* Set the TXP Mtu control register */
      nthw_field_set_val32(p->mpFldQueueMtuControlAdr, port);
    nthw_field_set_val32(p->mpFldQueueMtuControlCnt, 1);
    nthw_register_flush(p->mpRegQueueMtuControl, 1);

    /* Set the TXP Mtu control register */
    nthw_field_set_val32(p->mpFldQueueMtuDataMaxMtu, maxMtu);
    nthw_register_flush(p->mpRegQueueMtuData, 1);
  } else {
    NT_LOG(DBG, NTHW, "NthwEpp::%s - portType unsupported", __func__);
    nthw_register_reset(p->mpRegQueueMtuControl);
    nthw_register_flush(p->mpRegQueueMtuControl, 1);
    nthw_register_reset(p->mpRegQueueMtuData);
    nthw_register_flush(p->mpRegQueueMtuData, 1);
    nthw_register_reset(p->mpRegTxpPortMtuControl);
    nthw_register_flush(p->mpRegTxpPortMtuControl, 1);
    nthw_register_reset(p->mpRegTxpPortMtuData);
    nthw_register_flush(p->mpRegTxpPortMtuData, 1);

    return -ENOTSUP;
  }

  return 0;
}

int NthwEpp_setTxpQos(NthwEpp_t *p, uint32_t port, uint32_t informationRate, uint32_t informationRateFractional, uint32_t burstSize)
{
  if(p == NULL)
    return 0;

  nthw_field_set_val32(p->mpFldTxpQosControlAdr, port);
  nthw_field_set_val32(p->mpFldTxpQosControlCnt, 1);
  nthw_register_flush(p->mpRegTxpQosControl, 1);

  uint32_t enable = ((informationRate | informationRateFractional | burstSize) != 0);
  nthw_field_set_val32(p->mpFldTxpQosDataEnable, enable);
  nthw_field_set_val32(p->mpFldTxpQosDataInformationRate, informationRate);
  nthw_field_set_val32(p->mpFldTxpQosDataInformationRateFractional, informationRateFractional);
  nthw_field_set_val32(p->mpFldTxpQosDataBurstSize, burstSize);
  nthw_register_flush(p->mpRegTxpQosData, 1);

    return 0;
}

int NthwEpp_setVportQos(NthwEpp_t *p, uint32_t port, uint32_t informationRate, uint32_t informationRateFractional, uint32_t burstSize)
{
  if(p == NULL)
    return 0;

  nthw_field_set_val32(p->mpFldVportQosControlAdr, port);
  nthw_field_set_val32(p->mpFldVportQosControlCnt, 1);
  nthw_register_flush(p->mpRegVportQosControl, 1);

  uint32_t enable = ((informationRate | informationRateFractional | burstSize) != 0);
  nthw_field_set_val32(p->mpFldVportQosDataEnable, enable);
  nthw_field_set_val32(p->mpFldVportQosDataInformationRate, informationRate);
  nthw_field_set_val32(p->mpFldVportQosDataInformationRateFractional, informationRateFractional);
  nthw_field_set_val32(p->mpFldVportQosDataBurstSize, burstSize);
  nthw_register_flush(p->mpRegVportQosData, 1);

  return 0;
}

int NthwEpp_setQueueToVport(NthwEpp_t *p, uint32_t qid, uint32_t vport)
{
  if(p == NULL)
    return 0;

  nthw_field_set_val32(p->mpFldQueueVportControlAdr, qid);
  nthw_field_set_val32(p->mpFldQueueVportControlCnt, 1);
  nthw_register_flush(p->mpRegQueueVportControl, 1);

  nthw_field_set_val32(p->mpFldQueueVportDataVport, vport);
  nthw_register_flush(p->mpRegQueueVportData, 1);
  return 0;
}
