/* SPDX-License-Identifier: BSD-3-Clause
 * Copyright(c) 2023 Napatech A/S
 */

#include "ntlog.h"

#include "nthw_drv.h"
#include "nthw_register.h"

#include "nthw_mcu.h"

NthwMcu_t* NthwMcu_new(void)
{
  NthwMcu_t* p = malloc(sizeof(NthwMcu_t));
  if (p) {
    memset(p, 0, sizeof(NthwMcu_t));
  }
  return p;
}

void NthwMcu_delete(NthwMcu_t* p)
{
  if (p) {
    memset(p, 0, sizeof(NthwMcu_t));
    free(p);
  }
}

int NthwMcu_init(NthwMcu_t* p, nthw_fpga_t* pFpga, int nInstance)
{
  struct nthw_module_s* mod = Fpga_queryModule(pFpga, MOD_MCU, nInstance);

  if (p == NULL) {
    return mod == NULL ? -1 : 0;
  }

  if (mod == NULL) {
    NT_LOG(ERR, NTHW, "%s: MCU %d: no such instance\n", pFpga->p_fpga_info->mpAdapterIdStr, nInstance);
    return -1;
  }

  p->mpFpga = pFpga;
  p->mnInstance = nInstance;
  p->mpModMcu = mod;

#if defined(DEBUG) && (1)
  nthw_module_set_debug_mode(p->mpModMcu, 0xff); /* DEBUG: */
#endif

  p->mnModuleMajorVersion = nthw_module_get_major_version(p->mpModMcu);
  p->mnModuleMinorVersion = nthw_module_get_minor_version(p->mpModMcu);
  p->mnModuleVersionPacked64 = nthw_module_get_version_packed64(p->mpModMcu);

  p->mb_McuPresent = nthw_fpga_get_product_param(pFpga, NT_MCU_PRESENT, 0);
  p->mn_McuType = nthw_fpga_get_product_param(pFpga, NT_MCU_TYPE, -1);
  p->mn_McuDramSize = nthw_fpga_get_product_param(pFpga, NT_MCU_DRAM_SIZE, 0);

  p->mpRegCsr = nthw_module_get_register(p->mpModMcu, MCU_CSR);
  p->mpFldCsrHalt = nthw_register_get_field(p->mpRegCsr, MCU_CSR_HALT);
  p->mpFldCsrPause = nthw_register_get_field(p->mpRegCsr, MCU_CSR_PAUSE);
  p->mpFldCsrReset = nthw_register_get_field(p->mpRegCsr, MCU_CSR_RESET_MCU);

  p->mpRegDramCtrl = nthw_module_get_register(p->mpModMcu, MCU_DRAM_CTRL);
  p->mpFldDramCtrlAdr = nthw_register_get_field(p->mpRegDramCtrl, MCU_DRAM_CTRL_ADR);
  p->mpFldDramCtrlCnt = nthw_register_get_field(p->mpRegDramCtrl, MCU_DRAM_CTRL_CNT);

  p->mpRegDramRdData = nthw_module_get_register(p->mpModMcu, MCU_DRAM_RD_DATA);
  p->mpFldDramRdDataData = nthw_register_get_field(p->mpRegDramRdData, MCU_DRAM_RD_DATA_DATA);

  p->mpRegDramWrData = nthw_module_get_register(p->mpModMcu, MCU_DRAM_WR_DATA);
  p->mpFldDramWrDataData = nthw_register_get_field(p->mpRegDramWrData, MCU_DRAM_WR_DATA_DATA);

  p->mpRegIramCtrl = nthw_module_get_register(p->mpModMcu, MCU_IRAM_CTRL);
  p->mpFldIramCtrlAdr = nthw_register_get_field(p->mpRegIramCtrl, MCU_IRAM_CTRL_ADR);
  p->mpFldIramCtrlCnt = nthw_register_get_field(p->mpRegIramCtrl, MCU_IRAM_CTRL_CNT);

  p->mpRegIramData = nthw_module_get_register(p->mpModMcu, MCU_IRAM_DATA);
  p->mpFldIramDataData = nthw_register_get_field(p->mpRegIramData, MCU_IRAM_DATA_DATA);

  return 0;
}

void NthwMcu_setHalt(NthwMcu_t* p, uint32_t value)
{
  nthw_field_set_val_flush32(p->mpFldCsrHalt, value);
}

uint32_t NthwMcu_getHalt(NthwMcu_t* p)
{
  return nthw_field_get_val32(p->mpFldCsrHalt);
}

void NthwMcu_setPause(NthwMcu_t* p, uint32_t value)
{
  nthw_field_set_val_flush32(p->mpFldCsrPause, value);
}

uint32_t NthwMcu_getPause(NthwMcu_t* p)
{
  return nthw_field_get_val32(p->mpFldCsrPause);
}

void NthwMcu_setReset(NthwMcu_t* p, uint32_t value)
{
  nthw_field_set_val_flush32(p->mpFldCsrReset, value);
}

uint32_t NthwMcu_getReset(NthwMcu_t* p)
{
  return nthw_field_get_val32(p->mpFldCsrReset);
}

#if (0)
/* NOTE: currently not available */
int NthwMcu_IramRead(NthwMcu_t *p, uint32_t nAddr, uint8_t* pData, int nDataCnt)
{
  uint32_t* pData32 = (uint32_t*) pData;
  uint32_t nData32;
  int i;

  nthw_field_set_val32(p->mpFldIramCtrlAdr, nAddr);
  nthw_field_set_val_flush32(p->mpFldIramCtrlCnt, ((nDataCnt+1)>>2));

  for (i = 0; i < ((nDataCnt+1)>>2); i++) {
    pData32[i] = nthw_field_get_updated(p->mpFldIramDataData);
  }

  return -1;
}
#endif

int NthwMcu_IramWrite(NthwMcu_t *p, uint32_t nAddr, uint8_t* pData, int nDataCnt)
{
  uint32_t* pData32 = (uint32_t*) pData;
  int i;

  nthw_field_set_val32(p->mpFldIramCtrlAdr, nAddr);
  nthw_field_set_val_flush32(p->mpFldIramCtrlCnt, ((nDataCnt+1)>>2));

  for (i = 0; i < ((nDataCnt+1)>>2); i++) {
    nthw_field_set_val_flush32(p->mpFldIramDataData, pData32[i]);
  }

  return 0;
}

int NthwMcu_DramRead(NthwMcu_t *p, uint32_t nAddr, uint8_t* pData, int nDataCnt)
{
  uint32_t* pData32 = (uint32_t*) pData;
  int i;

  nthw_field_set_val32(p->mpFldDramCtrlAdr, nAddr);
  nthw_field_set_val_flush32(p->mpFldDramCtrlCnt, ((nDataCnt+1)>>2));

  for (i = 0; i < ((nDataCnt+1)>>2); i++) {
    pData32[i] = nthw_field_get_updated(p->mpFldDramRdDataData);
  }

  return 0;
}

int NthwMcu_DramWrite(NthwMcu_t *p, uint32_t nAddr, uint8_t* pData, int nDataCnt)
{
  uint32_t* pData32 = (uint32_t*) pData;
  int i;

  nthw_field_set_val32(p->mpFldDramCtrlAdr, nAddr);
  nthw_field_set_val_flush32(p->mpFldDramCtrlCnt, ((nDataCnt+1)>>2));

  for (i = 0; i < ((nDataCnt+1)>>2); i++) {
    nthw_field_set_val_flush32(p->mpFldDramWrDataData, pData32[i]);
  }
  return 0;
}
