#!/bin/bash

#
# Configuration
#
VENDOR_ID="18f4"
VENDOR_ID_HEX="0x${VENDOR_ID}"
PRODUCT_ID="051a"
NIC_DRIVER="vfio-pci"
NIC_AUTO=$(grep -l ${VENDOR_ID_HEX} /sys/bus/pci/devices/*.0/vendor | head -n1 | sed -E 's|/sys/bus/pci/devices/([0-9a-fA-F:.]+)/vendor|\1|')
VF_NUM="${vfCount:-4}"                      # use 4 VFs by default
NIC="${NicPci:-${NIC_AUTO}}"                # set default NT NIC to detected NT PCI ID
NIC_AUTO_INIT="${NicAutoInit:-disabled}"    # by default disable NT SmartNIC automatic configuration
RET_OK=0
RET_FAIL=1

#
# Functions
#
function nic_status() {
    echo "NT SmartNIC info:"
    echo "  NIC PCI address: ${NIC}"
    echo "  NIC type:        $(lspci -s ${NIC} | cut -d' ' -f2-)"
    echo "  NIC driver:      ${NIC_DRIVER}"
    echo "  VFs requested:   ${VF_NUM}"
    echo "  VFs configured:  $(cat /sys/bus/pci/devices/${NIC}/sriov_numvfs)"
}

function fail() {
    echo $*
    exit 1
}

function is_nic_auto_init_enabled() {
    if [ "${NIC_AUTO_INIT}" == "enabled" ] ; then
        return ${RET_OK}
    else
        return ${RET_FAIL}
    fi
}

function is_driver_loaded() {
    if [ -d "/sys/bus/pci/drivers/${NIC_DRIVER}" ] ; then
        return ${RET_OK}
    else
        return ${RET_FAIL}
    fi
}

function is_sriov_enabled() {
    if [ -f "/sys/bus/pci/drivers/${NIC_DRIVER}/module/parameters/enable_sriov" ] && \
        ( grep "^Y$" "/sys/bus/pci/drivers/${NIC_DRIVER}/module/parameters/enable_sriov" > /dev/null ) ; then
        return ${RET_OK}
    else
        return ${RET_FAIL}
    fi
}

function is_nic_bound_to_driver() {
    if [ -e /sys/bus/pci/devices/${NIC}/driver ] && \
        [ "${NIC_DRIVER}" == "$(basename $(readlink /sys/bus/pci/devices/${NIC}/driver))" ] ; then
        return ${RET_OK}
    else
        return ${RET_FAIL}
    fi
}

function is_vf_num_available() {
    tmp_configured_vfs=$(cat /sys/bus/pci/devices/${NIC}/sriov_numvfs)
    if [ ${tmp_configured_vfs} -ge ${VF_NUM} ] ; then
        return ${RET_OK}
    else
        return ${RET_FAIL}
    fi
}

function is_vf_bound_to_driver() {
    tmp_nic_root=$(echo ${NIC} | cut -d':' -f1-2)
    if ( ${DPDK_DEVBIND} -s | grep "^${tmp_nic_root}:.*${PRODUCT_ID}" | grep -v "drv=${NIC_DRIVER}" ) > /dev/null ; then
        return ${RET_FAIL}
    else
        return ${RET_OK}
    fi
}

function load_driver() {
    if ( ! is_nic_auto_init_enabled ) ; then
        return
    fi

    if ( ! is_driver_loaded ) ; then
        modprobe ${NIC_DRIVER} enable_sriov=1
    fi
}

function enable_sriov() {
    if ( ! is_nic_auto_init_enabled ) ; then
        return
    fi

    if ( ! is_sriov_enabled ) ; then
        echo "Y" > /sys/bus/pci/drivers/${NIC_DRIVER}/module/parameters/enable_sriov
    fi
}

function bind_nic_to_driver() {
    if ( ! is_nic_auto_init_enabled ) ; then
        return
    fi

    if ( ! is_nic_bound_to_driver ) ; then
        ${DPDK_DEVBIND} -b ${NIC_DRIVER} ${NIC}
    fi
}

function enable_vf_num() {
    if ( ! is_nic_auto_init_enabled ) ; then
        return
    fi

    if ( ! is_vf_num_available ) ; then
        tmp_max_vfs=$(cat /sys/bus/pci/devices/${NIC}/sriov_totalvfs)
        if [ ${VF_NUM} -gt ${tmp_max_vfs} ] ; then
            fail "ERROR: NIC ${NIC} supports up to ${tmp_max_vfs} VFs, but ${VF_NUM} VFs are required."
        fi
        if ( ! echo ${VF_NUM} > /sys/bus/pci/devices/${NIC}/sriov_numvfs 2>/dev/null ) ; then
            echo "INFO: Cannot configure ${VF_NUM} VFs. Re-binding NIC ${NIC} to ${NIC_DRIVER} and configuring VFs again."
            ${DPDK_DEVBIND} -u ${NIC}
            ${DPDK_DEVBIND} -b ${NIC_DRIVER} ${NIC}
            if ( ! echo ${VF_NUM} > /sys/bus/pci/devices/${NIC}/sriov_numvfs 2>/dev/null ) ; then
                echo "ERROR: Configuration of ${VF_NUM} VFs failed."
            fi
        fi
    fi
}

function assign_vf_to_driver() {
    if ( ! is_nic_auto_init_enabled ) ; then
        return
    fi

    # ignore error output in case that assignment was done before
    echo ${VENDOR_ID} ${PRODUCT_ID} > /sys/bus/pci/drivers/vfio-pci/new_id 2>/dev/null
}

#
# Main
#

# autodetect dpdk-devbind.py or fail
DPDK_DEVBIND=$(which dpdk-devbind.py 2>/dev/null)
if [ ! -e "${DPDK_DEVBIND}" ] ; then fail "ERROR: Cannot find dpdk-devbind.py."; fi

# check if detection was successful
if [ "$NIC" == "autodetect" ] ; then
    if [ "${NIC_AUTO}" == "" ] || [ ! -e /sys/bus/pci/devices/${NIC_AUTO}/vendor ] ; then fail "ERROR: Cannot find any NT SmartNIC, please configure its PCI address manually by NicPci parameter." ; fi
    NIC=${NIC_AUTO}
fi

# check if NIC exists
if [ "${NIC}" == "" ] || [ ! -d /sys/bus/pci/devices/${NIC} ] ; then fail "ERROR: Cannot find NIC ${NIC}." ; fi

# check if NIC is from Napatech
if ( ! grep -E "^${VENDOR_ID_HEX}$" "/sys/bus/pci/devices/${NIC}/vendor" > /dev/null ) ; then fail "ERROR: NIC ${NIC} is not supported." ; fi

# check if SRIOV is enabled
if [ "$(ls -A /sys/kernel/iommu_groups)" == "" ] ; then fail "ERROR: IOMMU is not enabled at boot time. Please update kernel boot options." ; fi

# load & check NIC driver
load_driver
if ( ! is_driver_loaded ) ; then fail "ERROR: NIC driver ${NIC_DRIVER} is not loaded." ; fi

# enable & check sriov
enable_sriov
if ( ! is_sriov_enabled ) ; then fail "ERROR: NIC driver ${NIC_DRIVER} doesn't support SRIOV." ; fi

# bind & check NIC to driver
bind_nic_to_driver
if ( ! is_nic_bound_to_driver ) ; then fail "ERROR: NIC ${NIC} is not bound to ${NIC_DRIVER}." ; fi

# configure & check number of VFs
enable_vf_num
if ( ! is_vf_num_available ) ; then fail "ERROR: Only $(cat /sys/bus/pci/devices/${NIC}/sriov_numvfs) of ${VF_NUM} VFs are configured at NIC ${NIC}." ; fi

# assign VFs to driver
assign_vf_to_driver

# check if ALL VFs are bound to vfio-pci
if ( ! is_vf_bound_to_driver ) ; then fail "ERROR: Some of NIC ${NIC} VFs are not bound to ${NIC_DRIVER}" ; fi

nic_status
echo "NT SmartNIC ${NIC} is configured."
